import argparse
import dataclasses
import inspect
from enum import Enum
from logging import getLogger
from typing import cast, ClassVar, Any, Optional, List, Type, Dict, Set, Union, Tuple

from .. import docstring, utils

# from ..utils import Dataclass, DataclassType
from .field_parsing import get_parsing_fn
from .field_metavar import get_metavar
from .wrapper import Wrapper

logger = getLogger(__name__)


class FieldWrapper(Wrapper[dataclasses.Field]):
    """
    The FieldWrapper class acts a bit like an 'argparse.Action' class, which
    essentially just creates the `option_strings` and `arg_options` that get
    passed to the `add_argument(*option_strings, **arg_options)` function of the
    `argparse._ArgumentGroup` (in this case represented by the `parent`
    attribute, an instance of the class `DataclassWrapper`).

    The `option_strings`, `required`, `help`, `metavar`, `default`, etc.
    attributes just autogenerate the argument of the same name of the
    above-mentioned `add_argument` function. The `arg_options` attribute fills
    in the rest and may overwrite these values, depending on the type of field.

    The `field` argument is the actually wrapped `dataclasses.Field` instance.
    """

    # Wether or not `simple_parsing` should add option_string variants where
    # underscores in attribute names are replaced with dashes.
    # For example, when set to `True`, "--no-cache" and "--no_cache" could both
    # be used to point to the same attribute `no_cache` on some dataclass.
    # TODO: This can often make "--help" messages a bit crowded
    add_dash_variants: ClassVar[bool] = False

    # Wether to add the `dest` to the list of option strings.
    add_dest_to_option_strings: ClassVar[bool] = True

    def __init__(self, field: dataclasses.Field, parent: Any = None, prefix: str = ""):
        super().__init__(wrapped=field, name=field.name)
        self.field: dataclasses.Field = field
        self.prefix: str = prefix
        self._parent: Any = parent
        # Holders used to 'cache' the properties.
        # (could've used cached_property with Python 3.8).
        self._option_strings: Optional[Set[str]] = None
        self._required: Optional[bool] = None
        self._docstring: docstring.AttributeDocString = docstring.AttributeDocString()
        self._help: Optional[str] = None
        self._metavar: Optional[str] = None
        self._default: Optional[Union[Any, List[Any]]] = None
        self._dest: Optional[str] = None
        # the argparse-related options:
        self._arg_options: Dict[str, Any] = {}
        self._dest_field: Optional["FieldWrapper"] = None
        self._type: Optional[Type[Any]] = None

        # stores the resulting values for each of the destination attributes.
        self._results: Dict[str, Any] = {}

    @property
    def arg_options(self) -> Dict[str, Any]:
        """Dictionary of values to be passed to the `add_argument` method.

        The main feature of this package is to infer these arguments
        automatically using features of the built-in `dataclasses` package, as
        well as Python's type annotations.

        By passing additional keyword arguments to the `field()`
        function, the autogenerated arguments can be overwriten,
        giving access to all of the usual argparse features know and love.

        NOTE: When passing an `action` keyword argument, we remove all the
        autogenerated options that aren't required by the Action class
        constructor.
        For example, when specifying a custom `action` like "store_true" or
        "store_false", the `type` argument autogenerated here shouldn't be
        passed to the constructor of the `argparse._StoreFalseAction`, so we
        discard it.
        """
        if self._arg_options:
            return self._arg_options
        # get the auto-generated options.
        options = self.get_arg_options()
        # overwrite the auto-generated options with given ones, if any.
        options.update(self.custom_arg_options)
        # only keep the arguments used by the Action constructor.
        action = options.get("action", "store")
        self._arg_options = only_keep_action_args(options, action)
        return self._arg_options

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: Any,
        option_string: Optional[str] = None,
    ):
        """Immitates a custom Action, which sets the corresponding value from
        `values` at the right destination in the `constructor_arguments` of the
        parser.

        TODO: Could be simplified by removing unused arguments, if we decide
        that there is no real value in implementing a CustomAction class.

        Args:
            parser (argparse.ArgumentParser): the `simple_parsing.ArgumentParser` used.
            namespace (argparse.Namespace): (unused).
            values (Any): The parsed values for the argument.
            option_string (Optional[str], optional): (unused). Defaults to None.
        """
        from simple_parsing import ArgumentParser

        parser = cast(ArgumentParser, parser)

        if self.is_reused:
            values = self.duplicate_if_needed(values)
            logger.debug(f"(replicated the parsed values: '{values}')")
        else:
            values = [values]

        self._results = {}

        for destination, value in zip(self.destinations, values):
            parent_dest, attribute = utils.split_dest(destination)
            value = self.postprocess(value)
            self._results[destination] = value
            parser.constructor_arguments[parent_dest][attribute] = value
            logger.debug(
                f"setting value of {value} in constructor arguments "
                f"of parent at key '{parent_dest}' and attribute "
                f"'{attribute}'"
            )

    def get_arg_options(self) -> Dict[str, Any]:
        """ Create the `parser.add_arguments` kwargs for this field. """
        if not self.field.init:
            # Don't add command-line arguments for fields that have `init=False`.
            # TODO: See (https://github.com/lebrice/SimpleParsing/issues/20#issue-619198492)
            return {}
        # TODO: Refactor this:
        # 1. Create a `get_argparse_options_for_field` function
        # 2. Use `get_argparse_options_for_annotation` below as part of that function
        # 3. Update the dict returned from 1. with values set in the field() function
        # 4. Update the dict from 3. with the values set by the DataclassWrapper, or
        # when this field is reused. (are they ever modified externally?)
        # 5. Return that dictionary.
        _arg_options: Dict[str, Any] = {}

        _arg_options["required"] = self.required
        _arg_options["dest"] = self.dest
        _arg_options["default"] = self.default
        _arg_options["metavar"] = get_metavar(self.type)

        if self.help:
            _arg_options["help"] = self.help
        elif self.default is not None:
            # issue 64: Need to add an empty 'help' string, so that the formatter
            # automatically adds the (default: '123')
            _arg_options["help"] = " "

        if utils.is_choice(self.field):
            _arg_options["type"] = str
            _arg_options["choices"] = list(self.choices)
            if utils.is_list(self.type):
                _arg_options["nargs"] = argparse.ZERO_OR_MORE
            # We use the default 'metavar' generated by argparse.
            _arg_options.pop("metavar", None)

        elif utils.is_optional(self.type):
            _arg_options["required"] = False

            type_arguments = utils.get_args(self.type)
            # NOTE: Optional[<something>] is always translated to
            # Union[<something>, NoneType]
            assert type_arguments
            non_none_types = [t for t in type_arguments if t is not type(None)]
            assert non_none_types

            if len(non_none_types) == 1:
                wrapped_type = non_none_types[0]
            else:
                wrapped_type = Union[tuple(non_none_types)]

            if utils.is_tuple(wrapped_type):
                # TODO: ISSUE 42: better handle optional/nested tuples.
                # For now we just assume that the item types are 'simple'.

                # IDEA: This chould probably be a static method that takes in the type
                # annotation, and uses a recursive call to fetch the arg options of the
                # item type, and uses some of the entries from that dict to construct
                # the arg options of the parent?
                # NOTE: We'd have to return different value for the `type` argument
                # depending on the nesting:
                # Tuple[int, int] -> <class int>
                # Tuple[int, str] -> <some autogenerated function>
                # Tuple[Tuple[int, str], Tuple[int, str]] -> <some autogenerated
                #   function *that doesn't just use `<class int> from above twice!
                #   (Since we want to support passing --foo '(a, 1)' '(b, 4)'
                # `)>
                _arg_options["type"] = get_parsing_fn(wrapped_type)
                _arg_options["nargs"] = utils.get_container_nargs(wrapped_type)

            elif utils.is_list(wrapped_type):
                _arg_options["type"] = utils.get_argparse_type_for_container(
                    wrapped_type
                )
                _arg_options["nargs"] = "*"
                # NOTE: Can't set 'const', since we'd get:
                # ValueError: nargs must be '?' to supply const
                # _arg_options["const"] = []
            else:
                _arg_options["type"] = get_parsing_fn(wrapped_type)
                # TODO: Should the 'nargs' really be '?' here?
                _arg_options["nargs"] = "?"
                # assert False, (wrapped_type, utils.is_tuple(wrapped_type))

        elif self.is_union:
            logger.debug("Parsing a Union type!")
            _arg_options["type"] = get_parsing_fn(self.type)

        elif self.is_enum:
            logger.debug(f"Adding an Enum attribute '{self.name}'")
            # we actually parse enums as string, and convert them back to enums
            # in the `process` method.
            logger.debug(f"self.choices = {self.choices}")
            _arg_options["choices"] = list(e.name for e in self.type)
            _arg_options["type"] = str
            # if the default value is an Enum, we convert it to a string.
            if self.default:

                def enum_to_str(e):
                    return e.name if isinstance(e, Enum) else e

                if self.is_reused:
                    _arg_options["default"] = [
                        enum_to_str(default) for default in self.default
                    ]
                else:
                    _arg_options["default"] = enum_to_str(self.default)

        elif self.is_list:
            logger.debug(f"Adding a List attribute '{self.name}': {self.type}")
            _arg_options["nargs"] = "*"

            if self.is_reused:
                # TODO: Only the 'single-level' lists (not reused) use the new
                # `get_parsing_fn` function (for now).
                type_fn = utils._parse_multiple_containers(self.type)
                type_fn.__name__ = utils.get_type_name(self.type)
                _arg_options["type"] = type_fn
            else:
                _arg_options["type"] = utils.get_argparse_type_for_container(self.type)

        elif utils.is_tuple(self.type):
            logger.debug(
                f"Adding a Tuple attribute '{self.name}' with type {self.type}"
            )
            _arg_options["nargs"] = utils.get_container_nargs(self.type)
            _arg_options["type"] = get_parsing_fn(self.type)

            if self.is_reused:
                type_fn = utils._parse_multiple_containers(self.type)
                type_fn.__name__ = utils.get_type_name(self.type)
                _arg_options["type"] = type_fn

        elif utils.is_bool(self.type):
            _arg_options["type"] = utils.str2bool
            _arg_options["type"].__name__ = "bool"
            _arg_options["nargs"] = "?"

        else:
            # "Plain" / simple argument.
            # For the metavar, use a custom passed value, if present, else do
            # not put any value in (which uses the usual value from argparse).
            if self.metavar:
                _arg_options["metavar"] = self.metavar
            else:
                # Remove the 'metavar' that we auto-generated above.
                _arg_options.pop("metavar", None)
            _arg_options["type"] = self.custom_arg_options.get(
                "type", get_parsing_fn(self.type)
            )

        if self.is_reused:
            if self.required:
                _arg_options["nargs"] = "+"
            else:
                _arg_options["nargs"] = "*"
        return _arg_options

    def duplicate_if_needed(self, parsed_values: Any) -> List[Any]:
        """Duplicates the passed argument values if needed, such that each instance gets a value.

        For example, if we expected 3 values for an argument, and a single value was passed,
        then we duplicate it so that each of the three instances get the same value.

        Args:
            parsed_values (Any): The parsed value(s)

        Raises:
            utils.InconsistentArgumentError: If the number of arguments passed is
            inconsistent (neither 1 or the number of instances)

        Returns:
            List[Any]: The list of parsed values, of the right length.
        """
        num_instances_to_parse = len(self.destinations)
        logger.debug(f"num to parse: {num_instances_to_parse}")
        logger.debug(f"(raw) parsed values: '{parsed_values}'")

        assert self.is_reused
        assert (
            num_instances_to_parse > 1
        ), "multiple is true but we're expected to instantiate only one instance"

        if utils.is_list(self.type) and isinstance(parsed_values, tuple):
            parsed_values = list(parsed_values)

        if not self.is_tuple and not self.is_list and isinstance(parsed_values, list):
            nesting_level = utils.get_nesting_level(parsed_values)
            if (
                nesting_level == 2
                and len(parsed_values) == 1
                and len(parsed_values[0]) == num_instances_to_parse
            ):
                return parsed_values[0]

        if not isinstance(parsed_values, (list, tuple)):
            parsed_values = [parsed_values]

        if len(parsed_values) == num_instances_to_parse:
            return parsed_values
        elif len(parsed_values) == 1:
            return parsed_values * num_instances_to_parse
        else:
            raise utils.InconsistentArgumentError(
                f"The field '{self.name}' contains {len(parsed_values)} values,"
                f" but either 1 or {num_instances_to_parse} values were "
                f"expected."
            )
        return parsed_values

    def postprocess(self, raw_parsed_value: Any) -> Any:
        """Applies any conversions to the 'raw' parsed value before it is used
        in the constructor of the dataclass.

        Args:
            raw_parsed_value (Any): The 'raw' parsed value.

        Returns:
            Any: The processed value
        """
        if self.is_enum:
            logger.debug(
                f"field postprocessing for Enum field '{self.name}' with value:"
                f" {raw_parsed_value}'"
            )
            if isinstance(raw_parsed_value, str):
                raw_parsed_value = self.type[raw_parsed_value]  # type: ignore
            return raw_parsed_value

        elif self.is_choice:
            choice_dict = self.field.metadata.get("choice_dict")
            if choice_dict:
                key_type = type(next(iter(choice_dict.keys())))
                if self.is_list and isinstance(raw_parsed_value[0], key_type):
                    return [choice_dict[value] for value in raw_parsed_value]
                elif isinstance(raw_parsed_value, key_type):
                    return choice_dict[raw_parsed_value]
            return raw_parsed_value

        elif self.is_tuple:
            logger.debug("we're parsing a tuple!")
            # argparse always returns lists by default. If the field was of a
            # Tuple type, we just transform the list to a Tuple.
            if not isinstance(raw_parsed_value, tuple):
                return tuple(raw_parsed_value)

        elif self.is_bool:
            # print(self.name, raw_parsed_value)
            if self.dest_field:
                # TODO: This isn't used anywhere, what was the idea again?
                # other_default = self.dest_field.field.metadata.get("_original_default")
                pass
                # print(other_default)

            if raw_parsed_value is None and self.default is not None:
                logger.debug("value is None, returning opposite of the default")
                return not self.default
            return raw_parsed_value

        elif self.is_list:
            if isinstance(raw_parsed_value, tuple):
                return list(raw_parsed_value)
            else:
                return raw_parsed_value

        elif self.is_subparser:
            return raw_parsed_value

        elif utils.is_optional(self.type):
            item_type = utils.get_args(self.type)[0]
            if utils.is_tuple(item_type) and isinstance(raw_parsed_value, list):
                # TODO: Make sure that this doesn't cause issues with NamedTuple types.
                return tuple(raw_parsed_value)

        elif self.type not in utils.builtin_types:
            # TODO: what if we actually got an auto-generated parsing function?
            try:
                # if the field has a weird type, we try to call it directly.
                return self.type(raw_parsed_value)
            except Exception as e:
                logger.debug(
                    f"Unable to instantiate the field '{self.name}' of type "
                    f"'{self.type}' by using the type as a constructor. "
                    f"Returning the raw parsed value instead "
                    f"({raw_parsed_value}, of type {type(raw_parsed_value)}). "
                    f"(Caught Exception: {e})"
                )
                return raw_parsed_value

        logger.debug(
            f"field postprocessing for field of type '{self.type}' and with "
            f"value '{raw_parsed_value}'"
        )
        return raw_parsed_value

    @property
    def is_reused(self) -> bool:
        return len(self.destinations) > 1

    @property
    def action(self) -> Union[str, Type[argparse.Action]]:
        """The `action` argument to be passed to `add_argument(...)`."""
        return self.custom_arg_options.get("action", "store")

    @property
    def action_str(self) -> str:
        if isinstance(self.action, str):
            return self.action
        return self.action.__name__

    @property
    def custom_arg_options(self) -> Dict[str, Any]:
        """Custom argparse options that overwrite those in `arg_options`.

        Can be set by using the `field` function, passing in a keyword argument
        that would usually be passed to the parser.add_argument(
        *option_strings, **kwargs) method.
        """
        return self.field.metadata.get("custom_args", {})

    @property
    def destinations(self) -> List[str]:
        return [
            f"{parent_dest}.{self.name}" for parent_dest in self.parent.destinations
        ]

    @property
    def option_strings(self) -> List[str]:
        """Generates the `option_strings` argument to the `add_argument` call.

        `parser.add_argument(*name_or_flags, **arg_options)`

        ## Notes:
        - Additional names for the same argument can be added via the `field`
        function.
        - Whenever the name of an attribute includes underscores ("_"), the same
        argument can be passed by using dashes ("-") instead. This also includes
        aliases.
        - If an alias contained leading dashes, either single or double, the
        same number of dashes will be used, even in the case where a prefix is
        added.

        For an illustration of this, see the aliases example.

        """

        dashes: List[str] = []  # contains the leading dashes.
        options: List[str] = []  # contains the name following the dashes.

        dash = "-" if len(self.name) == 1 else "--"
        option = f"{self.prefix}{self.name}"

        dashes.append(dash)
        options.append(option)

        if dash == "-":
            # also add a double-dash option:
            dashes.append("--")
            options.append(option)

        # add all the aliases that were passed to the `field` function.
        for alias in self.aliases:
            if alias.startswith("--"):
                dash = "--"
                name = alias[2:]
            elif alias.startswith("-"):
                dash = "-"
                name = alias[1:]
            else:
                dash = "-" if len(alias) == 1 else "--"
                name = alias
            option = f"{self.prefix}{name}"

            dashes.append(dash)
            options.append(option)

        # Additionally, add all name variants with the "_" replaced with "-".
        # For example, "--no-cache" will correctly set the `no_cache` attribute,
        # even if an alias isn't explicitly created.

        if FieldWrapper.add_dash_variants:
            additional_options = [
                option.replace("_", "-") for option in options if "_" in option
            ]
            additional_dashes = [
                "-" if len(option) == 1 else "--" for option in additional_options
            ]
            options.extend(additional_options)
            dashes.extend(additional_dashes)

        if type(self).add_dest_to_option_strings:
            dashes.append("-" if len(self.dest) == 1 else "--")
            options.append(self.dest)

        # remove duplicates by creating a set.
        option_strings = set(f"{dash}{option}" for dash, option in zip(dashes, options))
        # TODO: possibly sort the option strings, if argparse doesn't do it
        # already.
        return list(sorted(option_strings, key=len))

    # @property
    # def prefix(self) -> str:
    #     return self._prefix

    @property
    def aliases(self) -> List[str]:
        return self.field.metadata.get("alias", [])

    @property
    def dest(self) -> str:
        """Where the attribute will be stored in the Namespace."""
        self._dest = super().dest
        # TODO: If a custom `dest` was passed, and it is a `Field` instance,
        # find the corresponding FieldWrapper and use its `dest` instead of ours.
        if self.dest_field:
            self._dest = self.dest_field.dest
            self.custom_arg_options.pop("dest", None)
        return self._dest

    @property
    def is_proxy(self) -> bool:
        return self.dest_field is not None

    @property
    def dest_field(self) -> Optional["FieldWrapper"]:
        """Return the `FieldWrapper` for which `self` is a proxy (same dest).
        When a `dest` argument is passed to `field()`, and its value is a
        `Field`, that indicates that this Field is just a proxy for another.

        In such a case, we replace the dest of `self` with that of the other
        wrapper's we then find the corresponding FieldWrapper and use its `dest`
        instead of ours.
        """
        if self._dest_field is not None:
            return self._dest_field
        custom_dest = self.custom_arg_options.get("dest")
        if isinstance(custom_dest, dataclasses.Field):
            all_fields: List[FieldWrapper] = []
            for parent in self.lineage():
                all_fields.extend(parent.fields)  # type: ignore
            for other_wrapper in all_fields:
                if custom_dest is other_wrapper.field:
                    self._dest_field = other_wrapper
                    break
        return self._dest_field

    @property
    def nargs(self):
        return self.custom_arg_options.get("nargs", None)

    # @property
    # def const(self):
    #     return self.custom_arg_options.get("const", None)

    @property
    def default(self) -> Any:
        """Either a single default value, when parsing a single argument, or
        the list of default values, when this argument is reused multiple times
        (which only happens with the `ConflictResolution.ALWAYS_MERGE` option).

        In order of increasing priority, this could either be:
        1. The default attribute of the field
        2. the value of the corresponding attribute on the parent,
        if it has a default value
        """
        if self._default is not None:
            return self._default

        default: Any = utils.default_value(self.field)

        if default is dataclasses.MISSING:
            default = None

        if self.action == "store_true" and default is None:
            default = False
        if self.action == "store_false" and default is None:
            default = True

        if self.parent.defaults:
            # if the dataclass holding this field has a default value (either
            # when passed  manually or by nesting), use the corresponding
            # attribute on that default instance.

            # TODO: When that default value is 'None' (even for a dataclass),
            # then we need to.. ?

            defaults = []
            for default_dataclass_instance in self.parent.defaults:
                if default_dataclass_instance is None:
                    default_value = default
                else:
                    default_value = getattr(default_dataclass_instance, self.name)
                defaults.append(default_value)
            default = defaults[0] if len(defaults) == 1 else defaults

        if self.is_reused and default is not None:
            n_destinations = len(self.destinations)
            assert n_destinations >= 1
            if not isinstance(default, list) or len(default) != n_destinations:
                default = [default] * n_destinations
            assert len(default) == n_destinations, (
                f"Not the same number of default values and destinations. "
                f"(default: {default}, # of destinations: {n_destinations})"
            )

        self._default = default
        return self._default

    @default.setter
    def default(self, value: Any):
        self._default = value

    @property
    def required(self) -> bool:
        if self._required is not None:
            return self._required

        if self.action_str.startswith("store_"):
            # all the store_* actions do not require a value.
            self._required = False
        elif self.is_optional:
            self._required = False
        elif self.parent.required:
            # if the parent dataclass is required, then this attribute is too.
            # TODO: does that make sense though?
            self._required = True

        elif self.nargs in {"?", "*"}:
            self._required = False
        elif self.nargs == "+":
            self._required = True

        elif self.default is None:
            self._required = True
        elif self.is_reused:
            # if we're reusing this argument, the default value might be a list
            # of `MISSING` values.
            self._required = any(v == dataclasses.MISSING for v in self.default)
        else:
            self._required = False
        return self._required

    @required.setter
    def required(self, value: bool):
        self._required = value

    @property
    def type(self) -> Type[Any]:
        """Returns the wrapped field's type annotation."""
        if self._type is None:
            self._type = self.field.type
            
            if self.is_choice and self.choice_dict:
                if utils.is_list(self.field.type):
                    self._type = List[str]
                else:
                    self._type = str

        return self._type

    def __str__(self):
        return f"""<FieldWrapper for field '{self.dest}'>"""

    @property
    def is_choice(self) -> bool:
        return self.choices is not None

    @property
    def choices(self) -> Optional[List]:
        choices = self.custom_arg_options.get("choices")
        choice_dict = self.choice_dict
        if not choices:
            return None
        if len(choices) == 1 and isinstance(choices[0], dict):
            choice_dict = choices[0]
            assert False, "HERE"  # FIXME: Remove this, was debugging something?
            return choice_dict
        return choices

    @property
    def choice_dict(self) -> Optional[Dict]:
        return self.field.metadata.get("choice_dict")

    @property
    def help(self) -> Optional[str]:
        if self._help:
            return self._help
        try:
            self._docstring = docstring.get_attribute_docstring(
                self.parent.dataclass, self.field.name
            )
        except (SystemExit, Exception) as e:
            logger.debug(
                f"Couldn't find attribute docstring for field {self.name}, {e}"
            )
            self._docstring = docstring.AttributeDocString()

        if self._docstring.docstring_below:
            self._help = self._docstring.docstring_below
        elif self._docstring.comment_above:
            self._help = self._docstring.comment_above
        elif self._docstring.comment_inline:
            self._help = self._docstring.comment_inline
        return self._help

    @help.setter
    def help(self, value: str):
        self._help = value

    @property
    def metavar(self) -> Optional[str]:
        """Returns the 'metavar' when set using one of the `field` functions,
        else None.
        """
        if self._metavar:
            return self._metavar
        self._metavar = self.custom_arg_options.get("metavar")
        return self._metavar

    @metavar.setter
    def metavar(self, value: str):
        self._metavar = value

    @property
    def name(self) -> str:
        return self.field.name

    @property
    def is_list(self):
        return utils.is_list(self.type)

    @property
    def is_enum(self) -> bool:
        return utils.is_enum(self.type)

    @property
    def is_tuple(self) -> bool:
        return utils.is_tuple(self.type)

    @property
    def is_bool(self) -> bool:
        return utils.is_bool(self.type)

    @property
    def is_optional(self) -> bool:
        return utils.is_optional(self.field.type)

    @property
    def is_union(self) -> bool:
        return utils.is_union(self.field.type)

    @property
    def is_subparser(self) -> bool:
        return utils.is_subparser_field(self.field)

    @property
    def type_arguments(self) -> Optional[Tuple[Type, ...]]:
        return utils.get_type_arguments(self.type)

    @property
    def parent(self) -> "DataclassWrapper":
        return self._parent

    @property
    def subparsers_dict(self) -> Optional[Dict[str, Type]]:
        """The dict of subparsers, which is created either when using a
        Union[<dataclass_1>, <dataclass_2>] type annotation, or when using the
        `subparsers()` function.
        """
        if self.field.metadata.get("subparsers"):
            return self.field.metadata["subparsers"]
        elif self.is_union:
            type_arguments = utils.get_type_arguments(self.field.type)
            if type_arguments and any(map(utils.is_dataclass_type, type_arguments)):
                return {
                    utils.get_type_name(dataclass_type).lower(): dataclass_type
                    for dataclass_type in type_arguments
                }

    def add_subparsers(self, parser: argparse.ArgumentParser):
        assert self.is_subparser
        from simple_parsing import ArgumentParser  # Just for typing.

        # add subparsers for each dataclass type in the field.
        subparsers = parser.add_subparsers(
            title=self.name,
            description=self.help,
            dest=self.dest,
            parser_class=ArgumentParser,
        )
        subparsers.required = True
        for subcommand, dataclass_type in self.subparsers_dict.items():
            logger.debug(f"adding subparser '{subcommand}' for type {dataclass_type}")
            subparser = subparsers.add_parser(subcommand)
            # Just for typing correctness, as we didn't explicitly change
            # the return type of subparsers.add_parser method.)
            subparser = cast(ArgumentParser, subparser)
            subparser.add_arguments(dataclass_type, dest=self.dest)

    def equivalent_argparse_code(self):
        arg_options = self.arg_options.copy()
        arg_options_string = f"{{'type': {arg_options.pop('type', str).__qualname__}"
        arg_options_string += str(arg_options).replace("{", ", ")
        return f"group.add_argument(*{self.option_strings}, **{arg_options_string})"


def only_keep_action_args(
    options: Dict[str, Any], action: Union[str, Any]
) -> Dict[str, Any]:
    """Remove all the arguments in `options` that aren't required by the Action.

    Parameters
    ----------
    options : Dict[str, Any]
        A dictionary of options that would usually be passed to
        `add_arguments(*option_strings, **options)`.
    action : Union[str, Any]
        The action class or name.

    Returns
    -------
    Dict[str, Any]
        [description]
    """
    # TODO: explicitly test these custom actions?
    argparse_action_classes: Dict[str, Type[argparse.Action]] = {
        "store": argparse._StoreAction,
        "store_const": argparse._StoreConstAction,
        "store_true": argparse._StoreTrueAction,
        "store_false": argparse._StoreFalseAction,
        "append": argparse._AppendAction,
        "append_const": argparse._AppendConstAction,
        "count": argparse._CountAction,
        "help": argparse._HelpAction,
        "version": argparse._VersionAction,
        "parsers": argparse._SubParsersAction,
    }
    if action not in argparse_action_classes:
        # the provided `action` is not a standard argparse-action.
        # We don't remove any of the provided options.
        return options

    # Remove all the keys that aren't needed by the action constructor:
    action_class = argparse_action_classes[action]
    argspec = inspect.getfullargspec(action_class)

    if argspec.varargs is not None or argspec.varkw is not None:
        # if the constructor takes variable arguments, pass all the options.
        logger.debug("Constructor takes var args. returning all options.")
        return options

    args_to_keep = argspec.args + ["action"]

    kept_options, deleted_options = utils.keep_keys(options, args_to_keep)
    if deleted_options:
        logger.debug(
            f"Some auto-generated options were deleted, as they were "
            f"not required by the Action constructor: {deleted_options}."
        )
    if deleted_options:
        logger.debug(f"Kept options: \t{kept_options.keys()}")
        logger.debug(f"Removed options: \t{deleted_options.keys()}")
    return kept_options


def get_argparse_options_for_annotation(t: Type) -> Dict[str, Any]:
    """Generates some of the argparse arguments for `parser.add_argument` based on the
    given type annotation.

    This is able to generate some, but not all the necessary options.

    WIP: This is part of an attempt at refactoring things a little bit.
    """
    _arg_options: Dict[str, Any] = {}
    _arg_options["metavar"] = get_metavar(t)

    if utils.is_enum(t):
        _arg_options["type"] = str
        assert False, t.names
        _arg_options["choices"] = t.names
        if utils.is_list(t):
            _arg_options["nargs"] = argparse.ZERO_OR_MORE
        # We use the default 'metavar' generated by argparse.
        _arg_options.pop("metavar", None)

    elif utils.is_optional(t):
        type_arguments = utils.get_args(t)
        assert type_arguments
        if len(type_arguments) != 2:
            raise NotImplementedError(
                "Can only handle `Optional[<something>]`, not "
                "`Union[<something>, <something_else>, None]`"
            )
        # NOTE: Optional[<something>] is always translated to
        # Union[<something>, NoneType]
        wrapped_type: Type = type_arguments[0]

        # Recurse!
        options_for_wrapped_type = get_argparse_options_for_annotation(wrapped_type)
        _arg_options.update(options_for_wrapped_type)
        _arg_options["required"] = False
        _arg_options.setdefault("nargs", "?")
        # _arg_options.setdefault("default", None)

        # if utils.is_tuple(wrapped_type):
        #     # TODO: ISSUE 42: better handle optional/nested tuples.
        #     # For now we just assume that the item types are 'simple'.

        #     # IDEA: This chould probably be a static method that takes in the type
        #     # annotation, and uses a recursive call to fetch the arg options of the
        #     # item type, and uses some of the entries from that dict to construct
        #     # the arg options of the parent?
        #     # NOTE: We'd have to return different value for the `type` argument
        #     # depending on the nesting:
        #     # Tuple[int, int] -> <class int>
        #     # Tuple[int, str] -> <some autogenerated function>
        #     # Tuple[Tuple[int, str], Tuple[int, str]] -> <some autogenerated
        #     #   function *that doesn't just use `<class int> from above twice!
        #     #   (Since we want to support passing --foo '(a, 1)' '(b, 4)'
        #     # `)>
        #     _arg_options["type"] = get_parsing_fn(wrapped_type)
        #     _arg_options["nargs"] = utils.get_container_nargs(wrapped_type)

        # elif utils.is_list(wrapped_type):
        #     _arg_options["type"] = utils.get_argparse_type_for_container(
        #         wrapped_type
        #     )
        #     _arg_options["nargs"] = "*"
        #     # NOTE: Can't set 'const', since we'd get:
        #     # ValueError: nargs must be '?' to supply const
        #     # _arg_options["const"] = []
        # else:
        #     _arg_options["type"] = get_parsing_fn(wrapped_type)
        #     # TODO: Should the 'nargs' really be '?' here?
        #     _arg_options["nargs"] = "?"
        #     # assert False, (wrapped_type, utils.is_tuple(wrapped_type))

    elif utils.is_union(t):
        logger.debug("Parsing a Union type!")
        _arg_options["type"] = get_parsing_fn(t)

    elif utils.is_enum(t):
        # we actually parse enums as string, and convert them back to enums
        # in the `postprocess` method.
        assert False, t.names
        _arg_options["choices"] = list(e.name for e in t)
        _arg_options["type"] = str
        # if the default value is an Enum, we convert it to a string.
        # if self.default:
        #     def enum_to_str(e):
        #         return e.name if isinstance(e, Enum) else e

        #     if self.is_reused:
        #         _arg_options["default"] = [
        #             enum_to_str(default) for default in self.default
        #         ]
        #     else:
        #         _arg_options["default"] = enum_to_str(self.default)

    elif utils.is_list(t):
        logger.debug(f"Adding a List attribute '{t}")
        _arg_options["nargs"] = "*"
        item_type = utils.get_item_type(t)
        _arg_options["type"] = get_parsing_fn(item_type)

        # if self.is_reused:
        #     # TODO: Only the 'single-level' lists (not reused) use the new
        #     # `get_parsing_fn` function (for now).
        #     type_fn = utils._parse_multiple_containers(self.type)
        #     type_fn.__name__ = utils.get_type_name(self.type)
        #     _arg_options["type"] = type_fn
        # else:
        # _arg_options["type"] = utils.get_argparse_type_for_container(self.type)

    elif utils.is_tuple(t):
        logger.debug(f"Adding a Tuple attribute with type {t}")
        _arg_options["nargs"] = utils.get_container_nargs(t)
        _arg_options["type"] = get_parsing_fn(t)
        _arg_options["nargs"] = utils.get_container_nargs(t)

        # if self.is_reused:
        #     type_fn = utils._parse_multiple_containers(self.type)
        #     type_fn.__name__ = utils.get_type_name(self.type)
        #     _arg_options["type"] = type_fn

    elif utils.is_bool(t):
        _arg_options["type"] = utils.str2bool
        _arg_options["type"].__name__ = "bool"
        _arg_options["nargs"] = "?"

    else:
        # "Plain" / simple argument.
        # For the metavar, use a custom passed value, if present, else do
        # not put any value in (which uses the usual value from argparse).
        # _arg_options["metavar"] = get_metavar(t)
        _arg_options["type"] = get_parsing_fn(t)

    # if self.is_reused:
    #     if self.required:
    #         _arg_options["nargs"] = "+"
    #     else:
    #         _arg_options["nargs"] = "*"
    return _arg_options



import argparse
import dataclasses
import enum
from logging import getLogger
from typing import *
from typing import cast
from dataclasses import _MISSING_TYPE, MISSING
from .. import docstring, utils
from ..utils import Dataclass, DataclassType
from .wrapper import Wrapper
from .field_wrapper import FieldWrapper

logger = getLogger(__name__)

class DataclassWrapper(Wrapper[Dataclass]):
    def __init__(self,
                 dataclass: Type[Dataclass],
                 name: str,
                 default: Dataclass=None,
                 prefix: str="",
                 parent: "DataclassWrapper"=None,
                 _field: dataclasses.Field=None,
                 ):
        # super().__init__(dataclass, name)
        self.dataclass = dataclass
        self._name = name
        self.default = default 
        self.prefix = prefix

        self.fields: List[FieldWrapper] = []
        self._destinations: List[str] = []
        self._required: bool = False
        self._explicit: bool = False
        self._dest: str = ""
        self._children: List[DataclassWrapper] = []
        self._parent = parent
        # the field of the parent, which contains this child dataclass.
        self._field = _field

        # the default values
        self._defaults: List[Dataclass] = []

        if default:
            self.defaults = [default]

        self.optional: bool = False


        for field in dataclasses.fields(self.dataclass):
            if not field.init or field.metadata.get("cmd", True) == False:
                continue

            if utils.is_subparser_field(field) or utils.is_choice(field):
                wrapper = FieldWrapper(field, parent=self, prefix=prefix)
                self.fields.append(wrapper)
            
            elif utils.is_tuple_or_list_of_dataclasses(field.type):
                raise NotImplementedError(
                    f"Field {field.name} is of type {field.type}, which isn't "
                    f"supported yet. (container of a dataclass type)"
                )
            
            elif dataclasses.is_dataclass(field.type):
                # handle a nested dataclass attribute
                dataclass, name = field.type, field.name
                child_wrapper = DataclassWrapper(dataclass, name, parent=self, _field=field)
                self._children.append(child_wrapper)

            elif utils.contains_dataclass_type_arg(field.type):
                dataclass = utils.get_dataclass_type_arg(field.type)
                name = field.name
                child_wrapper = DataclassWrapper(dataclass, name, parent=self, _field=field, default=None)
                child_wrapper.required = False
                child_wrapper.optional = True
                self._children.append(child_wrapper)

            else:
                # a normal attribute
                field_wrapper = FieldWrapper(field, parent=self, prefix=self.prefix)
                logger.debug(f"wrapped field at {field_wrapper.dest} has a default value of {field_wrapper.default}")
                self.fields.append(field_wrapper)
        
        logger.debug(f"The dataclass at attribute {self.dest} has default values: {self.defaults}")


    def add_arguments(self, parser: argparse.ArgumentParser):
        from ..parsing import ArgumentParser
        parser = cast(ArgumentParser, parser)
        
        group = parser.add_argument_group(title=self.title, description=self.description)

        for wrapped_field in self.fields:
            if not wrapped_field.field.metadata.get("cmd", True):
                logger.debug(f"Skipping field {wrapped_field.name} because it has cmd=False.")
                continue

            if wrapped_field.is_subparser:
                wrapped_field.add_subparsers(parser)
                
            elif wrapped_field.arg_options:
                logger.debug(f"Arg options for field '{wrapped_field.name}': {wrapped_field.arg_options}")
                # TODO: CustomAction isn't very easy to debug, and is not working. Maybe look into that. Simulating it for now.
                group.add_argument(*wrapped_field.option_strings, **wrapped_field.arg_options)
    
    def equivalent_argparse_code(self, leading="group") -> str:
        code = ""
        import textwrap
        code += textwrap.dedent(f"""
        group = parser.add_argument_group(title="{self.title.strip()}", description="{self.description.strip()}")
        """)
        for wrapped_field in self.fields:
            if wrapped_field.is_subparser:
                # TODO:
                raise NotImplementedError("Subparsers equivalent is TODO.")
                code += textwrap.dedent(f"""\
                # add subparsers for each dataclass type in the field.
                subparsers = parser.add_subparsers(
                    title={wrapped_field.name},
                    description={wrapped_field.help},
                    dest={wrapped_field.dest},
                )
                subparsers.required = True

                for subcommand, dataclass_type in {self.subparsers_dict.items()}:
                    subparser = subparsers.add_parser(subcommand)
                    subparser = cast(ArgumentParser, subparser)
                    subparser.add_arguments(dataclass_type, dest=self.dest)
                """)
            elif wrapped_field.arg_options:
                code += textwrap.dedent(wrapped_field.equivalent_argparse_code()) + "\n"
        return code

    
    @property
    def name(self) -> str:
        return self._name

    @property
    def parent(self) -> Optional["DataclassWrapper"]:
        return self._parent

    @property
    def defaults(self) -> List[Dataclass]:
        if self._defaults:
            return self._defaults
        if self._field is None:
            return []
        assert self.parent is not None
        if self.parent.defaults:
            self._defaults = []
            for default in self.parent.defaults:
                if default is None:
                    default = None
                else:
                    default = getattr(default, self.name)
                self._defaults.append(default)
        else:
            default_field_value = utils.default_value(self._field)
            if isinstance(default_field_value, _MISSING_TYPE):
                self._defaults = []
            else:
                self._defaults = [default_field_value]
        return self._defaults

    @defaults.setter
    def defaults(self, value: List[Dataclass]):
        self._defaults = value

    @property
    def title(self) -> str:
        names_string = f""" [{', '.join(f"'{dest}'" for dest in self.destinations)}]"""
        title = self.dataclass.__qualname__ + names_string
        return title

    @property
    def description(self) -> str:
        if self.parent and self._field:    
            doc = docstring.get_attribute_docstring(self.parent.dataclass, self._field.name)            
            if doc is not None:
                if doc.docstring_below:
                    return doc.docstring_below
                elif doc.comment_above:
                    return doc.comment_above
                elif doc.comment_inline:
                    return doc.comment_inline
        return self.dataclass.__doc__ or ""

    # @property
    # def prefix(self) -> str:
    #     return self._prefix
    
    # @prefix.setter
    # def prefix(self, value: str):
    #     self._prefix = value
    #     for child_wrapper in self._children:
    #         child_wrapper.prefix = value

    @property
    def required(self) -> bool:
        return self._required

    @required.setter
    def required(self, value: bool):
        self._required = value
        for field in self.fields:
            field.required = value
        for child_wrapper in self._children:
            child_wrapper.required = value

    @property
    def multiple(self) -> bool:
        return len(self.destinations) > 1

    @property
    def descendants(self):
        for child in self._children:
            yield child
            yield from child.descendants

    @property
    def dest(self):
        lineage = []
        parent = self.parent
        while parent is not None:
            lineage.append(parent.name)
            parent = parent.parent
        lineage = list(reversed(lineage))
        lineage.append(self.name)
        _dest = ".".join(lineage)
        logger.debug(f"getting dest, returning {_dest}")
        return _dest
   

    @property
    def destinations(self) -> List[str]:
        if not self._destinations:
            if self.parent:
                self._destinations = [f"{d}.{self.name}" for d in self.parent.destinations]
            else:
                self._destinations = [self.name]
        return self._destinations

    @destinations.setter
    def destinations(self, value: List[str]):
        self._destinations = value

    def merge(self, other: "DataclassWrapper"):
        """Absorb all the relevant attributes from another wrapper.
        Args:
            other (DataclassWrapper): Another instance to absorb into this one.
        """
        # logger.debug(f"merging \n{self}\n with \n{other}")
        logger.debug(f"self destinations: {self.destinations}")
        logger.debug(f"other destinations: {other.destinations}")
        # assert not set(self.destinations).intersection(set(other.destinations)), "shouldn't have overlap in destinations"
        # self.destinations.extend(other.destinations)
        for dest in other.destinations:
            if dest not in self.destinations:
                self.destinations.append(dest)
        logger.debug(f"destinations after merge: {self.destinations}")
        self.defaults.extend(other.defaults)

        for field_wrapper in self.fields:
            field_wrapper.default = None

        for child, other_child in zip(self._children, other._children):
            child.merge(other_child)
        



""" Collection of helper classes and functions to reduce boilerplate code. """
from .fields import *
from .flatten import FlattenedAccess
from .serialization import Serializable, SimpleJsonEncoder, encode
from .hparams import HyperParameters

try:
    from .serialization import YamlSerializable
except ImportError:
    pass

# For backward compatibility purposes
JsonSerializable = Serializable
SimpleEncoder = SimpleJsonEncoder



import argparse
import enum
from abc import ABC, abstractmethod
from collections import defaultdict
from logging import getLogger
from typing import *
from . import utils
from .utils import Dataclass
from .wrappers import DataclassWrapper, FieldWrapper


logger = getLogger(__name__)


class ConflictResolution(enum.Enum):
    """Determines prefixing when adding the same dataclass more than once.

    
    - NONE:
        Dissallow using the same dataclass in two different destinations without
        explicitly setting a distinct prefix for at least one of them.

    - EXPLICIT:
        When adding arguments for a dataclass that is already present, the
        argparse arguments for each class will use their full absolute path as a
        prefix.

    - ALWAYS_MERGE:
        When adding arguments for a dataclass that has previously been added,
        the arguments for both the old and new destinations will be set using
        the same option_string, and the passed values for the old and new
        destinations will correspond to the first and second values,
        respectively.
        NOTE: This changes the argparse type for that argument into a list of
        the original item type.
    
    - AUTO (default):
        Prefixes for each destination are created automatically, using the first
        discriminative prefix that can differentiate between all the conflicting
        arguments.
    """
    NONE = -1
    EXPLICIT = 0
    ALWAYS_MERGE = 1
    AUTO = 2


class ConflictResolutionError(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Conflict(NamedTuple):
    option_string: str
    wrappers: List[FieldWrapper]


class ConflictResolver:
    def __init__(self, conflict_resolution = ConflictResolution.AUTO):
        self._wrappers: List[DataclassWrapper] = []
        self.conflict_resolution = conflict_resolution

    def resolve(self, wrappers: List[DataclassWrapper]) -> List[DataclassWrapper]:
        assert not self._wrappers
        for wrapper in wrappers:
            self._wrappers.append(wrapper)
            self._wrappers.extend(wrapper.descendants)
        
        conflict = self.get_conflict(self._wrappers)
        
        # current and maximum number of attemps. When reached, raises an error.
        cur_attempts, max_attempts = 0, 50
        while conflict:
            message: str = (
                "The following wrappers are in conflict, as they share the " +
                f"'{conflict.option_string}' option string:" +
                ("\n".join(str(w) for w in conflict.wrappers)) +
                f"(Conflict Resolution mode is {self.conflict_resolution})"
            )
            logger.debug(message)

            if self.conflict_resolution == ConflictResolution.NONE:           
                raise ConflictResolutionError(message)
                    
            elif self.conflict_resolution == ConflictResolution.EXPLICIT:
                self._fix_conflict_explicit(conflict)

            elif self.conflict_resolution == ConflictResolution.ALWAYS_MERGE:
                self._fix_conflict_merge(conflict)

            elif self.conflict_resolution == ConflictResolution.AUTO:
                self._fix_conflict_auto(conflict)

            conflict = self.get_conflict(self._wrappers)
            cur_attempts += 1
            if cur_attempts == max_attempts:
                raise ConflictResolutionError(
                    f"Reached maximum number of attempts ({max_attempts}) "
                    "while trying to solve the conflicting argument names. "
                    "This is either a bug, or there is something weird going "
                    "on with your class hierarchy/argument names... "
                    "\nIn any case, Please help us by submitting an issue on "
                    "the Github repo at "
                    "https://github.com/lebrice/SimpleParsing/issues, "
                    "or by using the following link: "
                    "https://github.com/lebrice/SimpleParsing/issues/new?"
                        "assignees=lebrice&"
                        "labels=bug"
                        "&template=bug_report.md"
                        "&title=BUG: ConflictResolutionError"
                    )

        assert not self._conflict_exists(self._wrappers)
        return self._wrappers

    def get_conflict(self, wrappers: Union[List[FieldWrapper], List[DataclassWrapper]]) -> Optional[Conflict]:
        field_wrappers: List[FieldWrapper] = []
        for w in wrappers:
            if isinstance(w, FieldWrapper):
                field_wrappers.append(w)
            else:
                field_wrappers.extend(w.fields)

        # TODO: #49: Also consider the conflicts with regular argparse arguments.

        conflicts: Dict[str, List[FieldWrapper]] = defaultdict(list)
        for field_wrapper in field_wrappers:
            for option_string in field_wrapper.option_strings:
                conflicts[option_string].append(field_wrapper)
        
        for option_string, wrappers in conflicts.items():
            if len(wrappers) > 1:
                return Conflict(option_string, wrappers)
        return None

    def _register(self, wrapper: Union[DataclassWrapper, FieldWrapper]) -> DataclassWrapper:
        if isinstance(wrapper, FieldWrapper):
            wrapper = wrapper.parent
        assert isinstance(wrapper, DataclassWrapper)
        logger.debug(f"Registering new DataclassWrapper: {wrapper}")
        self._wrappers.append(wrapper) 
        self._wrappers.extend(wrapper.descendants)       
        return wrapper

    def _unregister(self, wrapper: Union[DataclassWrapper, FieldWrapper]):
        if isinstance(wrapper, FieldWrapper):
            wrapper = wrapper.parent
        assert isinstance(wrapper, DataclassWrapper)
        logger.debug(f"Unregistering DataclassWrapper {wrapper}")
        self._wrappers.remove(wrapper)
        for child in wrapper.descendants:
            logger.debug(f"\tAlso Unregistering Child DataclassWrapper {child}")
            self._wrappers.remove(child)

    def _fix_conflict_explicit(self, conflict: Conflict):
        """Fixes conflicts between arguments following the "Explicit" approach.
        
        The Explicit approach gives a prefix to each argument which points to
        exactly where the argument is stored in the resulting Namespace. There
        can therefore not be any confusion between arguments, at the cost of
        having lengthy option strings.
                
        Parameters
        ----------
        - conflict : Conflict
        
            The conflict to hangle/fix.
        
        Raises
        ------
        ConflictResolutionError
            If its impossibe to fix the conflict.
        """
        logger.debug(f"fixing explicit conflict: {conflict}")
        
        if any(w.prefix for w in conflict.wrappers):
            raise ConflictResolutionError(
                "When in 'Explicit' mode, there shouldn't be a need for any user-set prefixes."
                "Just let the ArgumentParser set the explicit prefixes for all fields, and there won't be a conflict."
            )
        # TODO: Only set an explicit prefix on the fields that are in conflict
        
        # Check that there is no conflict between the fields after setting the explicit prefix.
        # If there is, that means that this conflict can't be fixed automatically, and a manual prefix should be set by the user.
        for field_wrapper in conflict.wrappers:
            explicit_prefix = field_wrapper.parent.dest + "."
            field_wrapper.prefix = explicit_prefix
        
        another_conflict = self.get_conflict(conflict.wrappers)
        if another_conflict and another_conflict.option_string == conflict.option_string:
            raise ConflictResolutionError(
                f"There is a conflict over the '{conflict.option_string}' "
                "option string, even after adding an explicit prefix to all "
                "the conflicting fields! \n"
                "To solve this, You can either use a different argument name, "
                "a different destination, or pass a differentiating prefix to "
                "`parser.add_arguments(<dataclass>, dest=destination, "
                "prefix=prefix)`"
            )
            

    def _fix_conflict_auto(self, conflict: Conflict):
        """Fixes a conflict using the AUTO method.
        
        Tries to find a discriminating prefix of minimal length for all the conflicting fields, using roughly the following pseudocode:

        1.  Sort the field wrappers by ascending nesting level.
            ("parent/root" wrappers first, children "leaf" wrappers last)
        2.  If the first wrapper is less nested than the others, remove it from the list (don't change its prefix)
        3.  For all the remaining wrappers, add one more "word" from their lineage (dest attribute) to their prefix,
            starting from the end and moving towards the parent.
        4.  If there is no conflict left, exit, else, return to step 1 with the new conflict.
            (This is performed implicitly by the method that calls this function, since it loops while there is a conflict). 
        
        Parameters
        ----------
        - conflict : Conflict
        
            The Conflict NamedTuple containing the conflicting option_string, as well as the conflicting `FieldWrapper`s.
        
        Raises
        ------
        ConflictResolutionError
            If its impossibe to fix the conflict.
        """
        field_wrappers = sorted(conflict.wrappers, key=lambda w: w.nesting_level)
        logger.debug(f"Conflict with options string '{conflict.option_string}':")
        for field in field_wrappers:
            logger.debug(f"Field wrapper: {field} nesting level: {field.nesting_level}.")

        assert len(set(field_wrappers)) >= 2, "Need at least 2 (distinct) FieldWrappers to have a conflict..."
        first_wrapper = field_wrappers[0]
        second_wrapper = field_wrappers[1]
        if first_wrapper.nesting_level < second_wrapper.nesting_level:
            # IF the first field_wrapper is a 'parent' of the following field_wrappers, then it maybe doesn't need an additional prefix.
            logger.debug(f"The first FieldWrapper is less nested than the others, removing it. ({first_wrapper})")
            field_wrappers.remove(first_wrapper)

        # add one more word to each of the remaining field_wrappers.
        for field_wrapper in field_wrappers:
            # Get the current and explicit (maximum) prefix:
            current_prefix = field_wrapper.prefix
            explicit_prefix = field_wrapper.parent.dest + "."

            logger.debug(f"current prefix: {current_prefix}, explicit prefix: {explicit_prefix}")
            if current_prefix == explicit_prefix:
                # We can't add any more words to the prefix of this FieldWrapper,
                # as it has already a prefix equivalent to its full destination...
                raise ConflictResolutionError(" ".join([
                    f"Cannot fix the conflict for the Options string {conflict.option_string},",
                    f"as the field {field_wrapper} already has the most explicit",
                    "prefix possible, and thus we can't add an additional",
                    "discriminating word to its prefix.",
                    "\n Consider modifying either the destination or the prefix",
                    "passed to `parser.add_arguments(<dataclass>, dest=destination, prefix=prefix)",
                ]))
            
            # find the next 'word' to add to the prefix.
            available_words = list(filter(bool, explicit_prefix.split(".")))
            used_words      = list(filter(bool,  current_prefix.split(".")))
            assert len(available_words) > len(used_words), "There should at least one word we haven't used yet!"
            logger.debug(f"Available words: {available_words}, used_words: {used_words}")
            
            n_available_words = len(available_words)
            n_used_words = len(used_words)
            word_to_add = available_words[(n_available_words-1) - n_used_words]
            logger.debug(f"Word to be added: {word_to_add}")
            field_wrapper.prefix = word_to_add + "." + current_prefix
            logger.debug(f"New prefix: {field_wrapper.prefix}")


    def _fix_conflict_merge(self, conflict: Conflict):
        """Fix conflicts using the merging approach.

        The first wrapper is kept, and the rest of the wrappers are absorbed
        into the first wrapper.
        
        # TODO: check that the ordering of arguments is still preserved!
        
        Parameters
        ----------
        conflict : Conflict
            The conflict NamedTuple.
        """
        fields = sorted(conflict.wrappers, key=lambda w: w.nesting_level)
        logger.debug(f"Conflict with options string '{conflict.option_string}':")
        for field in fields:
            logger.debug(f"Field wrapper: {field} nesting level: {field.nesting_level}.")

        assert len(conflict.wrappers) > 1
        first_wrapper: FieldWrapper = fields[0]
        self._unregister(first_wrapper.parent)

        for wrapper in conflict.wrappers[1:]:
            self._unregister(wrapper.parent)
            first_wrapper.parent.merge(wrapper.parent)
        
        assert first_wrapper.parent.multiple
        self._register(first_wrapper.parent)

    

    def _get_conflicting_group(self, all_wrappers: List[DataclassWrapper]) -> Optional[Conflict]:
        """Return the conflicting DataclassWrappers which share argument names.

        TODO: maybe return the list of fields, rather than the dataclasses?
        """
        conflicts: Dict[str, List[FieldWrapper]] = defaultdict(list)
        for wrapper in all_wrappers:
            for field in wrapper.fields:
                for option in field.option_strings:
                    conflicts[option].append(field)
        
        for option_string, fields in conflicts.items():
            if len(fields) > 1:
                # the dataclasses of the fields that share the same name.
                # wrappers: List[DataclassWrapper] = [f.parent for f in fields]
                # dataclasses = [wrapper.dataclass for wrapper in wrappers]
                # prefixes = [wrapper.prefix for wrapper in wrappers]
                # return Conflict(dataclasses[0], prefixes[0], wrappers)
                return Conflict(option_string, fields)
        return None

    def _conflict_exists(self, all_wrappers: List[DataclassWrapper]) -> bool:
        """Return True whenever a conflict exists. (option strings overlap). """
        arg_names: Set[str] = set()
        for wrapper in all_wrappers:
            for field in wrapper.fields:
                for option in field.option_strings:
                    if option in arg_names:
                        return True
                    arg_names.add(option)
        return False



""" Utility functions that simplify defining field of dataclasses. 
"""
import argparse
import dataclasses
import enum
import functools
import inspect
import json
import warnings
from collections import OrderedDict
from dataclasses import _MISSING_TYPE, MISSING
from enum import Enum
from typing import (Any, Callable, Dict, Iterable, List, Optional, Set, Tuple,
                    Type, TypeVar, Union, overload)

from simple_parsing.utils import (Dataclass, SimpleValueType,
                                  get_type_arguments, is_optional, is_tuple,
                                  is_union, str2bool)

from logging import getLogger

logger = getLogger(__name__)

E = TypeVar("E", bound=Enum)
K = TypeVar("K")
V = TypeVar("V")
T = TypeVar("T")


def field(default: Union[T, _MISSING_TYPE] = MISSING,
          alias: Optional[Union[str, List[str]]] = None,
          cmd: bool = True,
          *,
          to_dict: bool=True,
          encoding_fn: Callable[[T], Any]=None,
          decoding_fn: Callable[[Any], T]=None,
          # dataclasses.field arguments
          default_factory: Union[Callable[[], T], _MISSING_TYPE] = MISSING,
          init: bool = True,
          repr: bool = True,
          hash: Optional[bool] = None,
          compare: bool = True,
          metadata: Optional[Dict[str, Any]] = None,
          **custom_argparse_args: Any) -> T:
    """Extension of the `dataclasses.field` function.
    
    Adds the ability to customize how this field's command-line options are
    created, as well as how it is serialized / deseralized (if the containing 
    dataclass inherits from `simple_parsing.Serializable`.

    Leftover arguments are fed directly to the
    `ArgumentParser.add_argument(*option_strings, **kwargs)` method.

    Parameters
    ----------
    default : Union[T, _MISSING_TYPE], optional
        The default field value (same as in `dataclasses.field`), by default MISSING
    alias : Union[str, List[str]], optional
        Additional option_strings to pass to the `add_argument` method, by
        default None. When passing strings which do not start by "-" or "--", 
        will be prefixed with "-" if the string is one character and by "--"
        otherwise.
    cmd: bool, optional
        Wether to add command-line arguments for this field or not. Defaults to
        True.
    
    ## Serialization-related Keyword Arguments:

    to_dict : bool
        Wether to include this field in the dictionary when calling `to_dict()`.
        Defaults to True.
        Only has an effect when the dataclass containing this field is
        `Serializable`.
    encoding_fn : Callable[[T], Any], optional
        Function to apply to this field's value when encoding the dataclass to a
        dict. Only has an effect when the dataclass containing this field is
        `Serializable`.
    decoding_fn : Callable[[Any], T]. optional
        Function to use in order to recover a the value of this field from a
        serialized entry in a dictionary (inside `cls.from_dict`).
        Only has an effect when the dataclass containing this field is
        `Serializable`.
    
    ## Keyword Arguments of `dataclasses.field`

    default_factory : Union[Callable[[], T], _MISSING_TYPE], optional
        (same as in `dataclasses.field`), by default None
    init : bool, optional
        (same as in `dataclasses.field`), by default True
    repr : bool, optional
        (same as in `dataclasses.field`), by default True
    hash : bool, optional
        (same as in `dataclasses.field`), by default None
    compare : bool, optional
        (same as in `dataclasses.field`), by default True
    metadata : Dict[str, Any], optional
        (same as in `dataclasses.field`), by default None

    Returns
    -------
    T
        The value returned by the `dataclasses.field` function.
    """
    _metadata: Dict[str, Any] = metadata if metadata is not None else {}
    if alias:
        _metadata["alias"] = alias if isinstance(alias, list) else [alias]
    _metadata.update(dict(to_dict=to_dict))
    if encoding_fn is not None:
        _metadata.update(dict(encoding_fn=encoding_fn))
    if decoding_fn is not None:
        _metadata.update(dict(decoding_fn=decoding_fn))
    _metadata["cmd"] = cmd

    if custom_argparse_args:
        _metadata.update({"custom_args": custom_argparse_args})
        
        action = custom_argparse_args.get("action")
        if action == "store_false":
            if default not in {MISSING, True}:
                raise RuntimeError("default should either not be passed or set "
                                   "to True when using the store_false action.")
            default = True  # type: ignore
        
        elif action == "store_true":
            if default not in {MISSING, False}:
                raise RuntimeError("default should either not be passed or set "
                                   "to False when using the store_true action.")
            default = False  # type: ignore

    if default is not MISSING:
        return dataclasses.field(  # type: ignore
            default=default,
            init=init,
            repr=repr,
            hash=hash,
            compare=compare,
            metadata=_metadata
        )
    elif not isinstance(default_factory, dataclasses._MISSING_TYPE):
        return dataclasses.field(
            default_factory=default_factory,
            init=init,
            repr=repr,
            hash=hash,
            compare=compare,
            metadata=_metadata
        )
    else:
        return dataclasses.field(
            init=init,
            repr=repr,
            hash=hash,
            compare=compare,
            metadata=_metadata
        )


@overload
def choice(choices: Type[E], default: E, **kwargs) -> E:
    pass

@overload
def choice(choices: Dict[K, V], default: K, **kwargs) -> V:
    pass

def choice(*choices: T, default: Union[T, _MISSING_TYPE] = MISSING, **kwargs: Any) -> T:
    """ Makes a field which can be chosen from the set of choices from the
    command-line.

    Returns a regular `dataclasses.field()`, but with metadata which indicates  
    the allowed values.

    (New:) If `choices` is a dictionary, then passing the 'key' will result in
    the corresponding value being used. The values may be objects, for example.
    Similarly for Enum types, passing a type of enum will  

    Args:
        default (T, optional): The default value of the field. Defaults to dataclasses.MISSING,
        in which case the command-line argument is required.

    Raises:
        ValueError: If the default value isn't part of the given choices.

    Returns:
        T: the result of the usual `dataclasses.field()` function (a dataclass field/attribute).
    """
    assert len(choices) > 0, "Choice requires at least one positional argument!"

    if len(choices) == 1:
        choices = choices[0]
        if inspect.isclass(choices) and issubclass(choices, Enum):
            # If given an enum, construct a mapping from names to values.
            choice_enum: Type[Enum] = choices
            choices = OrderedDict(
                (e.name, e) for e in choice_enum
            )
            if default is not MISSING and not isinstance(default, choice_enum):
                if default in choices:
                    warnings.warn(UserWarning(
                        f"Setting default={default} could perhaps be ambiguous "
                        f"(enum names vs enum values). Consider using the enum "
                        f"value {choices[default]} instead."
                    ))
                    default = choices[default]
                else:
                    raise ValueError(f"'default' arg should be of type {choice_enum}, but got {default}")

        if isinstance(choices, dict):
            # if the choices is a dict, the options are the keys
            # save the info about the choice_dict in the field metadata.
            metadata = kwargs.setdefault("metadata", {})
            choice_dict = choices
            # save the choice_dict in metadata so that we can recover the values in postprocessing.
            metadata["choice_dict"] = choice_dict
            choices = list(choice_dict.keys())
        
            # TODO: If the choice dict is given, then add encoding/decoding functions that just
            # get/set the right key.
            def _encoding_fn(value: Any) -> str:
                """ Custom encoding function that will simply represent the value as the
                the key in the dict rather than the value itself.
                """
                if value in choice_dict.keys():
                    return value
                elif value in choice_dict.values():
                    return [k for k, v in choice_dict.items() if v == value][0]
                return value        
            kwargs.setdefault("encoding_fn", _encoding_fn)
            
            def _decoding_fn(value: Any) -> str:
                """ Custom decoding function that will retrieve the value from the
                stored key in the dictionary.
                """
                return choice_dict.get(value, value)
    
            kwargs.setdefault("decoding_fn", _decoding_fn)

            
    return field(
        default=default,
        choices=choices,
        **kwargs
    )


def list_field(*default_items: T, **kwargs) -> List[T]:
    """shorthand function for setting a `list` attribute on a dataclass,
    so that every instance of the dataclass doesn't share the same list.

    Accepts any of the arguments of the `dataclasses.field` function.

    Returns:
        List[T]: a `dataclasses.field` of type `list`, containing the `default_items`. 
    """
    default = kwargs.pop("default", None)
    if isinstance(default, list):
        # can't have that. field wants a default_factory.
        # we just give back a copy of the list as a default factory,
        # but this should be discouraged.
        from copy import deepcopy
        kwargs["default_factory"] = lambda: deepcopy(default)
    return mutable_field(list, default_items, **kwargs)


def dict_field(default_items: Union[Dict[K, V], Iterable[Tuple[K, V]]]=None, **kwargs) -> Dict[K, V]:
    """shorthand function for setting a `dict` attribute on a dataclass,
    so that every instance of the dataclass doesn't share the same `dict`.

    NOTE: Do not use keyword arguments as you usually would with a dictionary
    (as in something like `dict_field(a=1, b=2, c=3)`). Instead pass in a
    dictionary instance with the items: `dict_field(dict(a=1, b=2, c=3))`.
    The reason for this is that the keyword arguments are interpreted as custom
    argparse arguments, rather than arguments of the `dict` function!) 

    Also accepts any of the arguments of the `dataclasses.field` function.

    Returns:
        Dict[K, V]: a `dataclasses.Field` of type `Dict[K, V]`, containing the `default_items`. 
    """
    if default_items is None:
        default_items = {}
    elif isinstance(default_items, dict):
        default_items = default_items.items()
    return mutable_field(dict, default_items, **kwargs)


def set_field(*default_items: T, **kwargs) -> Set[T]:
    return mutable_field(set, default_items, **kwargs)


def mutable_field(_type: Type[T], *args, init: bool = True, repr: bool = True, hash: bool = None, compare: bool = True, metadata: Dict[str, Any] = None, **kwargs) -> T:
    # TODO: Check wether some of the keyword arguments are destined for the `field` function, or for the partial?    
    default_factory = kwargs.pop("default_factory", functools.partial(_type, *args))
    return field(default_factory=default_factory, init=init, repr=repr, hash=hash, compare=compare, metadata=metadata, **kwargs)


MutableField = mutable_field


def subparsers(subcommands: Dict[str, Type[Dataclass]], **kwargs) -> Any:
    return field(metadata={
        "subparsers": subcommands,
    }, **kwargs)


def flag(default: bool, **kwargs):
    """ Creates a boolean field with a default value of `default` and nargs='?'.
    """
    action = "store_true" if default is False else "store_false"
    return field(default=default, nargs="?", action=action, type=str2bool, **kwargs)



"""Simple, Elegant Argument parsing.
@author: Fabrice Normandin
"""
from . import helpers, utils, wrappers
from .conflicts import ConflictResolution
from .help_formatter import SimpleHelpFormatter
from .helpers import (MutableField, Serializable, choice, field, flag,
                      list_field, mutable_field, subparsers)
from .parsing import ArgumentParser, ParsingError
from .utils import InconsistentArgumentError

__all__ = [
    "helpers", "utils", "wrappers",
    "ConflictResolution",
    "MutableField", "Serializable", "SimpleHelpFormatter", "choice",
    "field", "flag", "list_field", "mutable_field", "subparsers",
    "ArgumentParser",
    "InconsistentArgumentError",
]

from . import _version
__version__ = _version.get_versions()['version']



"""Abstact Wrapper base-class for the FieldWrapper and DataclassWrapper."""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Generic, Optional, List, Type

from ..utils import T

class Wrapper(Generic[T], ABC):
    def __init__(self, wrapped: T, name: str):
        self.wrapped = wrapped
        self._dest: Optional[str] = None
        

    @property
    def dest(self) -> str:
        """ Where the attribute will be stored in the Namespace. """
        lineage_names: List[str] = [w.name for w in self.lineage()]
        self._dest = ".".join(reversed([self.name] + lineage_names))
        assert self._dest is not None
        return self._dest

    def lineage(self) -> List["Wrapper"]:
        lineage: List[Wrapper] = []
        parent = self.parent
        while parent is not None:
            lineage.append(parent)
            parent = parent.parent
        return lineage   

    @property
    def nesting_level(self) -> int:
        return len(self.lineage())
        level = 0
        parent = self.parent
        while parent is not None:
            parent = parent.parent
            level += 1
        return level
    
    @abstractmethod
    def equivalent_argparse_code(self) -> str:
        pass
    
    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @property
    @abstractmethod
    def parent(self) -> Optional["Wrapper"]:
        pass